"""
This module contains all functions that belong to graphics handling
and drawing on the game screen.
"""

from .. import gamelib
from .pifiles import map_color_profiles

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from .picross import PicrossGame

SELECTED_THEME = "BORED_GREY"
"""
Can select from the following themes:

'BORED_GREY': The default one, the most 'meh'.
"""

PROFILE = map_color_profiles()[SELECTED_THEME]

NLGS_LOGO = {                                                                                                                                                            (5, 0) : PROFILE["GOLD"],          (6, 0) : PROFILE["GOLD"],          (7, 0) : PROFILE["GOLD"],
                                                                                                       (3, 1) : PROFILE["GOLD"],        (4, 1) : PROFILE["GOLD"],        (5, 1) : PROFILE["DARK_GREY_3"],   (6, 1) : PROFILE["DARK_GREY_3"],   (7, 1) : PROFILE["DARK_GREY_3"],   (8, 1) : PROFILE["GOLD"],         (9, 1) : PROFILE["GOLD"],
                                                                       (2, 2) : PROFILE["GOLD"],       (3, 2) : PROFILE["BLACK"],       (4, 2) : PROFILE["BLACK"],       (5, 2) : PROFILE["BLACK"],         (6, 2) : PROFILE["BLACK"],         (7, 2) : PROFILE["DARK_GREY_3"],   (8, 2) : PROFILE["DARK_GREY_3"],  (9, 2) : PROFILE["DARK_GREY_3"],  (10, 2) : PROFILE["GOLD"],
                                       (1, 3) : PROFILE["GOLD"],       (2, 3) : PROFILE["BLACK"],      (3, 3) : PROFILE["BLACK"],       (4, 3) : PROFILE["BLACK"],       (5, 3) : PROFILE["BLACK"],         (6, 3) : PROFILE["BLACK"],         (7, 3) : PROFILE["BLACK"],         (8, 3) : PROFILE["BLACK"],        (9, 3) : PROFILE["DARK_GREY_3"],  (10, 3) : PROFILE["DARK_GREY_3"], (11, 3) : PROFILE["GOLD"],
                                       (1, 4) : PROFILE["GOLD"],       (2, 4) : PROFILE["BLACK"],      (3, 4) : PROFILE["SKIN_COLOR"],  (4, 4) : PROFILE["SKIN_COLOR"],  (5, 4) : PROFILE["SKIN_COLOR"],    (6, 4) : PROFILE["BLACK"],         (7, 4) : PROFILE["SKIN_COLOR"],    (8, 4) : PROFILE["DARK_GREY_3"],  (9, 4) : PROFILE["BLACK"],        (10, 4) : PROFILE["DARK_GREY_3"], (11, 4) : PROFILE["GOLD"],
             (0, 5) : PROFILE["GOLD"], (1, 5) : PROFILE["BLACK"],      (2, 5) : PROFILE["SKIN_COLOR"], (3, 5) : PROFILE["SKIN_COLOR"],  (4, 5) : PROFILE["SKIN_COLOR"],  (5, 5) : PROFILE["BLACK"],         (6, 5) : PROFILE["SKIN_COLOR"],    (7, 5) : PROFILE["SKIN_COLOR"],    (8, 5) : PROFILE["DARK_GREY_3"],  (9, 5) : PROFILE["DARK_GREY_3"],  (10, 5) : PROFILE["DARK_GREY_3"], (11, 5) : PROFILE["DARK_GREY_3"], (12, 5) : PROFILE["GOLD"],
             (0, 6) : PROFILE["GOLD"], (1, 6) : PROFILE["SKIN_COLOR"], (2, 6) : PROFILE["SKIN_COLOR"], (3, 6) : PROFILE["SKIN_COLOR"],  (4, 6) : PROFILE["DARK_BROWN"],  (5, 6) : PROFILE["SKIN_COLOR"],    (6, 6) : PROFILE["DARK_BROWN"],    (7, 6) : PROFILE["SKIN_COLOR"],    (8, 6) : PROFILE["DARK_GREY_3"],  (9, 6) : PROFILE["DARK_GREY_3"],  (10, 6) : PROFILE["DARK_GREY_3"], (11, 6) : PROFILE["DARK_GREY_3"], (12, 6) : PROFILE["GOLD"],
             (0, 7) : PROFILE["GOLD"], (1, 7) : PROFILE["SKIN_COLOR"], (2, 7) : PROFILE["SKIN_COLOR"], (3, 7) : PROFILE["SKIN_COLOR"],  (4, 7) : PROFILE["DARK_BROWN"],  (5, 7) : PROFILE["SKIN_COLOR"],    (6, 7) : PROFILE["DARK_BROWN"],    (7, 7) : PROFILE["SKIN_COLOR"],    (8, 7) : PROFILE["DARK_GREY_3"],  (9, 7) : PROFILE["DARK_GREY_3"],  (10, 7) : PROFILE["DARK_GREY_3"], (11, 7) : PROFILE["DARK_GREY_3"], (12, 7) : PROFILE["GOLD"],
                                       (1, 8) : PROFILE["GOLD"],       (2, 8) : PROFILE["SKIN_COLOR"], (3, 8) : PROFILE["SKIN_COLOR"],  (4, 8) : PROFILE["SKIN_COLOR"],  (5, 8) : PROFILE["SKIN_COLOR"],    (6, 8) : PROFILE["SKIN_COLOR"],    (7, 8) : PROFILE["SKIN_COLOR"],    (8, 8) : PROFILE["DARK_GREY_3"],  (9, 8) : PROFILE["DARK_GREY_3"],  (10, 8) : PROFILE["DARK_GREY_3"], (11, 8) : PROFILE["GOLD"],
                                       (1, 9) : PROFILE["GOLD"],       (2, 9) : PROFILE["SKIN_COLOR"], (3, 9) : PROFILE["SKIN_COLOR"],  (4, 9) : PROFILE["CANDY_RED"],   (5, 9) : PROFILE["CANDY_RED"],     (6, 9) : PROFILE["SKIN_COLOR"],    (7, 9) : PROFILE["SKIN_COLOR"],    (8, 9) : PROFILE["DARK_GREY_3"],  (9, 9) : PROFILE["DARK_GREY_3"],  (10, 9) : PROFILE["DARK_GREY_3"], (11, 9) : PROFILE["GOLD"],
                                                                       (2, 10) : PROFILE["GOLD"],      (3, 10) : PROFILE["SKIN_COLOR"], (4, 10) : PROFILE["SKIN_COLOR"], (5, 10) : PROFILE["SKIN_COLOR"],   (6, 10) : PROFILE["SKIN_COLOR"],   (7, 10) : PROFILE["SKIN_COLOR"],   (8, 10) : PROFILE["DARK_GREY_3"], (9, 10) : PROFILE["DARK_GREY_3"], (10, 10) : PROFILE["GOLD"],
                                                                                                       (3, 11) : PROFILE["GOLD"],       (4, 11) : PROFILE["GOLD"],       (5, 11) : PROFILE["SKIN_COLOR"],   (6, 11) : PROFILE["DARK_GREY_3"],  (7, 11) : PROFILE["DARK_GREY_3"],  (8, 11) : PROFILE["GOLD"],        (9, 11) : PROFILE["GOLD"],
                                                                                                                                                                         (5, 12) : PROFILE["GOLD"],         (6, 12) : PROFILE["GOLD"],         (7, 12) : PROFILE["GOLD"]
            }
"""
The legendary NLGS logo.
"""

def draw_GUI(game: 'PicrossGame') -> None:
    """
    Draws the Graphic User Interface of the game.

    It includes some info texts, and the counters used to display some game attributes.
    """

    gamelib.draw_rectangle(0, 0, game.width + 10, game.gui_size, outline=PROFILE["DARK_GREY_2"], fill=PROFILE["DARK_GREY_1"])

    gamelib.draw_text(f"Current Mistakes: {game.mistakes_count} / {game.max_mistakes}", (game.width / 50), (game.width / 25), size=(game.width // 33), anchor='w', fill=PROFILE["WHITE"])

    actual_grid_x, actual_grid_y = game.dimensions()

    if (not game.grid_x == (actual_grid_x - 1)) or (not game.grid_y == (actual_grid_y - 1)): # The actual size also counts the indexes

        gamelib.draw_text(f"Selected values differ!\nPress '{game.new_game.msg}'\n <= to apply",
                          110, 35, size=(game.width // 50), justify="left", anchor="nw", fill=PROFILE["WHITE"])

    gamelib.draw_text("Grid Dimensions:", (game.width * 0.69), (game.width / 25), size=(game.width // 33), anchor='c', fill=PROFILE["WHITE"])

    cols_x, cols_y = (game.width * 0.65), (game.gui_size * 0.65)
    rows_x, rows_y = (game.width * 0.73), (game.gui_size * 0.65)

    gamelib.draw_text("Cols (x)", (game.width * 0.51), cols_y, size=(game.width // 33), anchor='c', fill=PROFILE["WHITE"])
    gamelib.draw_text("Rows (y)", (game.width * 0.88), rows_y, size=(game.width // 33), anchor='c', fill=PROFILE["WHITE"])

    frame_margin = int(game.gui_size / 6.666)

    gamelib.draw_rectangle(cols_x - frame_margin, cols_y - frame_margin, cols_x + frame_margin, cols_y + frame_margin, width=(game.width // 200), outline=PROFILE["DARK_GREY_2"], fill=PROFILE["DARK_GREY_3"])
    gamelib.draw_rectangle(rows_x - frame_margin, rows_y - frame_margin, rows_x + frame_margin, rows_y + frame_margin, width=(game.width // 200), outline=PROFILE["DARK_GREY_2"], fill=PROFILE["DARK_GREY_3"])

    gamelib.draw_text(f"{game.grid_x}", cols_x, cols_y, size=(game.width // 50), anchor='c', fill=PROFILE["WHITE"])
    gamelib.draw_text(f"{game.grid_y}", rows_x, rows_y, size=(game.width // 50), anchor='c', fill=PROFILE["WHITE"])

def draw_buttons(game: 'PicrossGame'):
    """
    Draws all the buttons of the games, in their respective coordinates.
    """

    for button in game.buttons_list:

        gamelib.draw_rectangle(button.x1, button.y1, button.x2, button.y2, width=(game.width // 250), outline=PROFILE["DARK_GREY_5"], fill=PROFILE["DARK_GREY_4"], activefill=(PROFILE["GREY_2"] if button.msg else ''))

        center_x, center_y = button.center()
        if button.msg: gamelib.draw_text(button.msg, center_x, center_y, size=(game.width // 50), justify='c', anchor='c', fill=PROFILE["WHITE"])

def draw_megaphone(game: 'PicrossGame') -> None:
    """
    Draws a megaphone that shows if the game has audio or not.
    """

    if game.has_audio:

        gamelib.draw_arc((game.width * 0.91),    # x1
                         (game.gui_size * 0.05), # y1
                         (game.width * 0.98),    # x2
                         (game.gui_size * 0.35), # y2
                         start="135", outline='', fill=PROFILE["WHITE"])

        gamelib.draw_arc((game.width * 0.916),   # x1
                         (game.gui_size * 0.05), # y1
                         (game.width * 0.986),   # x2
                         (game.gui_size * 0.35), # y2
                         start="135", outline='', fill=PROFILE["DARK_GREY_4"])

        gamelib.draw_arc((game.width * 0.924),   # x1
                         (game.gui_size * 0.08), # y1
                         (game.width * 0.972),   # x2
                         (game.gui_size * 0.32), # y2
                         start="135", outline='', fill=PROFILE["WHITE"])

        gamelib.draw_arc((game.width * 0.93),    # x1
                         (game.gui_size * 0.05), # y1
                         (game.width * 0.98),    # x2
                         (game.gui_size * 0.35), # y2
                         start="135", outline='', fill=PROFILE["DARK_GREY_4"])

    else:

        gamelib.draw_line((game.width * 0.91),    # x1
                          (game.gui_size * 0.15), # y1
                          (game.width * 0.93),    # x2
                          (game.gui_size * 0.25), # y2
                          width=(game.gui_size // 50), fill=PROFILE["WHITE"])

        gamelib.draw_line((game.width * 0.93),    # x1
                          (game.gui_size * 0.15), # y1
                          (game.width * 0.91),    # x2
                          (game.gui_size * 0.25), # y2
                          width=(game.gui_size // 50), fill=PROFILE["WHITE"])

        gamelib.draw_line((game.width * 0.97),    # x1
                          (game.gui_size * 0.05), # y1
                          (game.width * 0.93),    # x2
                          (game.gui_size * 0.35), # y2
                          width=int(game.gui_size * 0.03))

    gamelib.draw_polygon([(game.width * 0.944),   # x1
                         (game.gui_size * 0.07),  # y1
                         (game.width * 0.944),    # x2
                         (game.gui_size * 0.33),  # y2
                         (game.width * 0.98),     # x3
                         (game.gui_size * 0.20)], # y3
                         outline='', fill=PROFILE["WHITE"])

    gamelib.draw_rectangle((game.width * 0.97),    # x1
                           (game.gui_size * 0.15), # y1
                           (game.width * 0.986),   # x2
                           (game.gui_size * 0.25), # y2
                           outline=PROFILE["DARK_GREY_2"], fill=PROFILE["WHITE"])

def draw_grid(game: 'PicrossGame') -> None:
    """
    Draws the grid of the game, according to the window sizes and grid columns
    and rows.
    """

    cols, rows = game.dimensions()

    index_x, index_y = game.calculate_indexes()

    cell_x = (game.width - index_x) / (cols - 1)
    cell_y = (game.height - index_y - game.gui_size) / (rows - 1)

    error_x = cell_x / 4
    error_y = cell_y / 4

    gamelib.draw_rectangle(0, game.gui_size, game.width + 10, game.height + 10, outline='', fill=PROFILE["GREY_1"])

    for j in range(1, rows):

        for i in range(1, cols):

            cell = game.grid[j][i]

            cell_color = (PROFILE["LIGHT_GREY_2"] if cell[1] else (PROFILE["BLACK"] if cell[0] else PROFILE["WHITE"]))

            gamelib.draw_rectangle(((i - 1) * cell_x) + index_x,
                                ((j - 1) * cell_y) + game.gui_size + index_y,
                                (i * cell_x) + index_x,
                                (j * cell_y) + game.gui_size + index_y,
                                outline='', fill=cell_color, activefill=(PROFILE["LIGHT_GREY_1"] if cell[1] else ''))

            if (i, j) in game.mistakes_coord:

                gamelib.draw_line(((i - 1) * cell_x) + index_x + error_x,
                                ((j - 1) * cell_y) + game.gui_size + index_y + error_y,
                                (i * cell_x) + index_x - error_x,
                                (j * cell_y) + game.gui_size + index_y - error_y,
                                width=(game.width // 150), fill=PROFILE["RED"])

                gamelib.draw_line((i * cell_x) + index_x - error_x,
                                ((j - 1) * cell_y) + game.gui_size + index_y + error_y,
                                ((i - 1) * cell_x) + index_x + error_x,
                                (j * cell_y) + game.gui_size + index_y - error_y,
                                width=(game.width // 150), fill=PROFILE["RED"])

    for x in range(cols + 1):

        gamelib.draw_line((x * cell_x) + index_x, game.gui_size, (x * cell_x) + index_x, game.height, width=(game.width // 500), fill=PROFILE["ALMOST_BLACK"])

    for y in range(rows + 1):

        gamelib.draw_line(0, (y * cell_y) + game.gui_size + index_y, game.width, (y * cell_y) + game.gui_size + index_y, width=(game.height // 700), fill=PROFILE["ALMOST_BLACK"])

def draw_index_numbers(game: 'PicrossGame') -> None:
    """
    Writes in the screen the numbers used for the indexes of the game grid.
    """

    cols, rows = game.dimensions()

    index_x, index_y = game.calculate_indexes()

    cell_x = (game.width - index_x) / (cols - 1)
    cell_y = (game.height - index_y - game.gui_size) / (rows - 1)

    for j in range(1, rows):

        row_text = "  ".join([str(number) for number in game.grid[j][0]])
        row_text_size = int(index_x //  ((game.width / 200) * len(game.grid[j][0]) + 1))

        gamelib.draw_text(row_text,
                         (index_x / 2),
                         (j * cell_y) - (cell_y / 2) + game.gui_size + index_y,
                         size=row_text_size, justify='c', anchor='c', fill=PROFILE["WHITE"])

    for i in range(1, cols):

        col_text = "\n\n".join([str(number) for number in game.grid[0][i]])
        col_text_size = int(index_y // ((game.gui_size / 33.333) * len(game.grid[0][i]) + 1))

        gamelib.draw_text(col_text,
                         (i * cell_x) - (cell_x / 2) + index_x,
                         ((index_y / 2) + game.gui_size),
                         size=col_text_size, justify='c', anchor='c', fill=PROFILE["WHITE"])

def draw_figure(x1: int, y1: int, x2: int, y2: int, grid_cols: int, grid_rows: int, colors_dict: dict) -> None:
    """
    ______________________________________________________________________

    x1, y1, x2, y2: <int>

    grid_cols, grid_rows: <int>

    colors_dict: <dict> --> {<tuple> : <str>, <tuple> : <str>, ... , <tuple> : <str>}
    --> {(<int>, <int>) : <str>, (<int>, <int>) : <str>, ... , (<int>, <int>) : <str>}


    ---> None
    ______________________________________________________________________

    Given an area to play with, the grid sizes into which the space is divided,
    and a dictionary with all colors to fill, this function draws any pixel art as
    it draws a color rectangle on any given coordinates inside the area and designated
    grid.
    """

    cell_x = (x2 - x1) / grid_cols
    cell_y = (y2 - y1) / grid_rows

    for x in range(grid_cols):

        for y in range(grid_rows):

            gamelib.draw_rectangle((x * cell_x) + x1,
                                   (y * cell_y) + y1,
                                   ((x + 1) * cell_x) + x1,
                                   ((y + 1) * cell_y) + y1,
                                   outline='', fill=(colors_dict.get((x, y), '')))

def draw_screen(game: 'PicrossGame') -> None:
    """
    Draws the entirety of the game elements on the screen.
    """

    draw_grid(game)
    draw_GUI(game)
    draw_buttons(game)

    draw_megaphone(game)

    draw_index_numbers(game)

    index_x, index_y = game.calculate_indexes()
    draw_figure(0, game.gui_size, index_x, index_y + game.gui_size, 13, 13, NLGS_LOGO)


"""
Absolute paths of anything that the program needs.
"""

from importlib.resources import path as fpath
from typing import Optional

def file_path(filename: str, subpackage: Optional[str]=None) -> str:
    """
    Returns the absolute path of a file associated with a package or subpackage.
    """

    subpackage_path = f"gamelibgames{f'.{subpackage}' if subpackage else ''}"

    return str(fpath(subpackage_path, filename))


PICROSS_ICON = file_path("picross_logo.gif", "picross.img.icon")
PICROSS_PROFILES = file_path("color_profiles.txt", "picross")
PI_TILE_WRONG = file_path("tile_wrong.wav", "picross.sfx")
PI_TILE_RIGHT = file_path("tile_wrong.wav", "picross.sfx")
PI_GAME_WON = file_path("game_won.wav", "picross.sfx")
PI_GAME_LOST = file_path("game_lost.wav", "picross.sfx")

MINE_ICON = file_path("minesweeper_logo.gif", "minesweeper.img.icon")
MINE_FIGURES =  file_path("figures.txt", "minesweeper")
MINE_PROFILES = file_path("color_profiles.txt", "minesweeper")
MINE_MARKED = file_path("marked.wav", "minesweeper.sfx")
MINE_CELL = file_path("uncovered_{cell}.wav", "minesweeper.sfx")
MINE_BTN_UP = file_path("button_up.wav", "minesweeper.sfx")
MINE_BTN_DN = file_path("button_dn.wav", "minesweeper.sfx")
MINE_POPUP = file_path("popup.wav", "minesweeper.sfx")
MINE_BOOM = file_path("boom.wav", "minesweeper.sfx")
MINE_VICTORY = file_path("victory.wav", "minesweeper.sfx")
MINE_NEW_GAME = file_path("new_game.wav", "minesweeper.sfx")


from .. import gamelib
from . import minefiles

from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from .minesweeper import MinesweeperGame

THEME = "DEFAULT"
"""
Can select between:

* DEFAULT - The vey common one
"""

PROFILE = minefiles.map_color_profiles()[THEME]

FIGURES = minefiles.map_figures()

def draw_GUI(game: 'MinesweeperGame') -> None:
    """
    Draws all of the elements of the Graphic User Interface.
    """

    gamelib.draw_rectangle(0, 0, game.width + 50, game.gui_size, width=(game.width // 300), outline=PROFILE["GUI_OUTLINE_1"], fill=PROFILE["GUI_1"])

    gamelib.draw_rectangle(game.width * 0.65,
                           0,
                           game.width * 0.75,
                           game.gui_size,
                           outline=PROFILE["GUI_OUTLINE_1"], fill=PROFILE["GUI_2"])

    gamelib.draw_text(f"X\n{game.grid_x}",
                      game.width * 0.7,
                      (game.gui_size / 2),
                      justify='c', anchor='c',
                      size=(game.width // 40),
                      fill=PROFILE["GUI_TEXT"])

    gamelib.draw_rectangle(game.width * 0.85,
                           0,
                           game.width * 0.95,
                           game.gui_size,
                           outline=PROFILE["GUI_OUTLINE_1"], fill=PROFILE["GUI_2"])

    gamelib.draw_text(f"Y\n{game.grid_y}",
                      game.width * 0.9,
                      (game.gui_size / 2),
                      justify='c', anchor='c',
                      size=(game.width // 40),
                      fill=PROFILE["GUI_TEXT"])

    gamelib.draw_rectangle(game.width * 0.25,
                           0,
                           game.width * 0.35,
                           game.gui_size,
                           outline=PROFILE["GUI_OUTLINE_1"], fill=PROFILE["GUI_2"])

    gamelib.draw_text(f"Mines:\n{game.mines if game.mines > 0 else 'Auto'}",
                      game.width * 0.3,
                      (game.gui_size / 2),
                      justify='c', anchor='c',
                      size=(game.width // 40),
                      fill=PROFILE["GUI_TEXT"])

def draw_grid(game: 'MinesweeperGame') -> None:
    """
    Draws the game grid.
    """

    cols, rows = game.dimensions()

    extra_space = game.cell_size / 24

    for i in range(cols + 1):

        gamelib.draw_line(game.cell_size * i,
                          game.gui_size,
                          game.cell_size * i,
                          game.height,
                          width=(game.cell_size // 10), fill=PROFILE["CELL_COLOR_3"])

    for j in range(rows + 1):

        gamelib.draw_line(0,
                          (game.cell_size * j) + game.gui_size,
                          game.width,
                          (game.cell_size * j) + game.gui_size,
                          width=(game.cell_size // 10), fill=PROFILE["CELL_COLOR_3"])

    for y in range(rows):

        for x in range(cols):

            cell = game.grid[y][x]

            gamelib.draw_rectangle((game.cell_size * x) + extra_space,
                                   (game.cell_size * y) + game.gui_size + extra_space,
                                   (game.cell_size * (x + 1)) - extra_space,
                                   (game.cell_size * (y + 1)) + game.gui_size - extra_space,
                                   outline=PROFILE["CELL_OUTLINE"], fill=PROFILE["CELL_COLOR_1" if cell is not game.loser_cell else "LOSER_CELL"])

            if cell.is_hidden:

                gamelib.draw_rectangle(game.cell_size * x,
                                      (game.cell_size * y) + game.gui_size,
                                      game.cell_size * (x + 1) - extra_space,
                                      (game.cell_size * y) + game.gui_size + (extra_space * 1.5),
                                      outline=PROFILE["CELL_OUTLINE"], fill=PROFILE["CELL_COLOR_2"])

                gamelib.draw_rectangle(game.cell_size * x,
                                      (game.cell_size * y) + game.gui_size,
                                      (game.cell_size * x) + (extra_space * 1.5),
                                      (game.cell_size * (y + 1)) + game.gui_size - extra_space,
                                      outline=PROFILE["CELL_OUTLINE"], fill=PROFILE["CELL_COLOR_2"])

                if cell.state == 1: # is flagged

                    draw_flag(game, x, y)

                elif cell.state == 2: # dunno

                    draw_dunno(game, x, y)

            else:

                if cell.value == 'X':

                    draw_mine(game, x, y)

                elif not cell.value in (0, ''):

                    # gamelib.draw_text(str(cell.value), (game.cell_size * (x + 0.5)), (game.cell_size * (y + 0.5)) + game.gui_size, justify='c', anchor='c', fill=PROFILE[f"NUMBER_{cell.value}"])
                    draw_number(game, x, y, cell.value)

def draw_flag(game: 'MinesweeperGame', x: int, y: int) -> None:
    """
    Draws the flag if the indicated cell is flagged.
    """

    extra_space = game.cell_size / 10

    # Post
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 2),
                           (game.cell_size * (y + 1)) + game.gui_size - (extra_space * 2.5),
                           (game.cell_size * (x + 1)) - (extra_space * 2),
                           (game.cell_size * (y + 1)) + game.gui_size - extra_space,
                           outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_2"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 3.5),
                          (game.cell_size * (y + 1)) + game.gui_size - (extra_space * 3),
                          (game.cell_size * (x + 1)) - (extra_space * 3.5),
                          (game.cell_size * (y + 1)) + game.gui_size - (extra_space * 2.5),
                          outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_2"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 4.5),
                          (game.cell_size * (y + 0.5)) + game.gui_size,
                          (game.cell_size * (x + 1)) - (extra_space * 4.5),
                          (game.cell_size * (y + 1)) + game.gui_size - (extra_space * 3),
                          outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_2"])

    # Flag
    gamelib.draw_rectangle((game.cell_size * (x + 0.5)) - extra_space,
                           (game.cell_size * y) + game.gui_size + (extra_space * 2),
                           (game.cell_size * (x + 1)) - (extra_space * 4.5),
                           (game.cell_size * (y + 0.5)) + game.gui_size,
                           outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_1"])

    gamelib.draw_rectangle((game.cell_size * (x + 0.5)) - (extra_space * 2.5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 2.5),
                           (game.cell_size * (x + 0.5)) - extra_space,
                           (game.cell_size * (y + 0.5)) + game.gui_size - (extra_space * 0.8),
                           outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 1.5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 3),
                           (game.cell_size * (x + 0.5)) - (extra_space * 2.5),
                           (game.cell_size * (y + 0.5)) + game.gui_size - (extra_space * 1.3),
                           outline=PROFILE["FLAG_OUTLINE"], fill=PROFILE["FLAG_1"])

def draw_dunno(game: 'MinesweeperGame', x: int, y: int) -> None:
    """
    Draws a question mark to show that a cell has said symbol.
    """

    extra_space = game.cell_size / 13

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 3),
                           (game.cell_size * y) + game.gui_size + (extra_space * 3),
                           (game.cell_size * x) + (extra_space * 5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 4),
                           (game.cell_size * y) + game.gui_size + (extra_space * 2),
                           (game.cell_size * x) + (extra_space * 9),
                           (game.cell_size * y) + game.gui_size + (extra_space * 3),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                           (game.cell_size * y) + game.gui_size + (extra_space * 3),
                           (game.cell_size * x) + (extra_space * 10),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                           (game.cell_size * x) + (extra_space * 8),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                           (game.cell_size * x) + (extra_space * 8),
                           (game.cell_size * y) + game.gui_size + (extra_space * 9),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 10),
                           (game.cell_size * x) + (extra_space * 8),
                           (game.cell_size * y) + game.gui_size + (extra_space * 12),
                            outline=PROFILE["DUNNO_OUTLINE"], fill=PROFILE["DUNNO_1"])

def draw_mine(game: 'MinesweeperGame', x: int, y: int) -> None:
    """
    Draws a mine if the cell indicated by the parameter coordiantes has it.
    """

    extra_space = game.cell_size / 20

    # Transversal Lines - Vertical
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 9.5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 3),
                           (game.cell_size * x) + (extra_space * 10.5),
                           (game.cell_size * (y + 1)) + game.gui_size - (extra_space * 3),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Transversal Lines - Horizontal
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 3),
                           (game.cell_size * y) + game.gui_size + (extra_space * 9.5),
                           (game.cell_size * (x + 1)) - (extra_space * 3),
                           (game.cell_size * y) + game.gui_size + (extra_space * 10.5),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Major Body
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                           (game.cell_size * x) + (extra_space * 15),
                           (game.cell_size * y) + game.gui_size + (extra_space * 13),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Upper Details
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                           (game.cell_size * x) + (extra_space * 14),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                           (game.cell_size * y) + game.gui_size + (extra_space * 5),
                           (game.cell_size * x) + (extra_space * 13),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Bottom Details
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 13),
                           (game.cell_size * x) + (extra_space * 14),
                           (game.cell_size * y) + game.gui_size + (extra_space * 14),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                           (game.cell_size * y) + game.gui_size  + (extra_space * 14),
                           (game.cell_size * x) + (extra_space * 13),
                           (game.cell_size * y) + game.gui_size  + (extra_space * 15),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Upper-Left Particle
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 5),
                           (game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Upper-Right Particle
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                           (game.cell_size * y) + game.gui_size + (extra_space * 5),
                           (game.cell_size * x) + (extra_space * 15),
                           (game.cell_size * y) + game.gui_size + (extra_space * 6),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Bottom-Left Particle
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                           (game.cell_size * y) + game.gui_size + (extra_space * 14),
                           (game.cell_size * x) + (extra_space * 6),
                           (game.cell_size * y) + game.gui_size + (extra_space * 15),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Bottom-Right Particle
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                           (game.cell_size * y) + game.gui_size + (extra_space * 14),
                           (game.cell_size * x) + (extra_space * 15),
                           (game.cell_size * y) + game.gui_size + (extra_space * 15),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_1"])

    # Reflection
    gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                           (game.cell_size * y) + game.gui_size + (extra_space * 7),
                           (game.cell_size * x) + (extra_space * 9),
                           (game.cell_size * y) + game.gui_size + (extra_space * 9),
                           outline=PROFILE["MINE_OUTLINE"], fill=PROFILE["MINE_2"])

def draw_number(game: 'MinesweeperGame', x: int, y: int, number: int) -> None:
    """
    Draws a little pixel art of the number that is the value of the cell.
    """

    number_color = PROFILE[f"NUMBER_{number}"]
    extra_space = game.cell_size / 20

    if number == 1:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13.5),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 9),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9.5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                               (game.cell_size * y) + game.gui_size + (extra_space * 8.5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7.5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 8.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 9),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6.5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 10),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5.5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6.5),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 2:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 12),
                               (game.cell_size * x) + (extra_space * 9),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 11),
                               (game.cell_size * y) + game.gui_size + (extra_space * 12),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 10),
                               (game.cell_size * x) + (extra_space * 13),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 10),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 10),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 11),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                               (game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 8),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 13),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 3:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 4:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 11),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                               (game.cell_size * x) + (extra_space * 9),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 7),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 10),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 5:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                               (game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 10),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 6:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                               (game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 10),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 7:

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 11),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 10),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 13),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

        gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 9),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

    elif number == 8:

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 5),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 7),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 8),
                               (game.cell_size * y) + game.gui_size + (extra_space * 13),
                               (game.cell_size * x) + (extra_space * 12),
                               (game.cell_size * y) + game.gui_size + (extra_space * 15),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                               (game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 6),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 9),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 5),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 6),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

       gamelib.draw_rectangle((game.cell_size * x) + (extra_space * 14),
                               (game.cell_size * y) + game.gui_size + (extra_space * 11),
                               (game.cell_size * x) + (extra_space * 15),
                               (game.cell_size * y) + game.gui_size + (extra_space * 14),
                                outline=PROFILE["NUMBER_OUTLINE"], fill=number_color)

def draw_buttons(game: 'MinesweeperGame') -> None:
    """
    Draws all the buttons of the game.
    """

    for button in game.buttons:

        gamelib.draw_rectangle(button.x1,
                               button.y1,
                               button.x2,
                               button.y2,
                               width=(game.cell_size // 15),
                               outline=PROFILE["BUTTON_OUTLINE" if button is not game.new_game else "GUI_OUTLINE_1"],
                               fill=PROFILE["BUTTON_1" if button is not game.new_game else "NEW_GAME_1"],
                               activefill=PROFILE["BUTTON_2" if button is not game.new_game else "NEW_GAME_2"])

        if button.msg:

            cx, cy = button.center()

            gamelib.draw_text(button.msg,
                            cx,
                            cy,
                            size=(game.cell_size // 2),
                            justify='c', anchor='c',
                            fill=PROFILE["BUTTON_TEXT"])

    x1, y1, x2, y2 = game.new_game.all_coord()

    gamelib.draw_rectangle(x1,
                           y1,
                           x1 + (game.width * 0.004166667),
                           y2 - (game.gui_size * 0.02),
                           outline=PROFILE["CELL_OUTLINE"], fill=PROFILE["CELL_COLOR_2"])

    gamelib.draw_rectangle(x1,
                           y1,
                           x2 - (game.width * 0.002),
                           y1 + (game.gui_size * 0.08),
                           outline=PROFILE["CELL_OUTLINE"], fill=PROFILE["CELL_COLOR_2"])

def draw_counter(game: 'MinesweeperGame', x1: int, y1: int, x2: int, y2: int) -> None:
    """
    Draws the flag counter which tells how much flags should be left
    (and, if used correctly, aldo how many mines there is still).

    It can enter in negative values. This is intentional as it tells
    that the player has placed more flags than mines there should be
    on the grid.
    """

    gamelib.draw_rectangle(0,
                           0,
                           game.width * 0.15,
                           game.gui_size,
                           outline=PROFILE["GUI_OUTLINE_1"], fill=PROFILE["GUI_3"])

    digits = str(game.flags)
    digits_amount = len(digits)

    if digits_amount > 3:

        space_per_digit = (x2 - x1) / digits_amount

    else:

        space_per_digit = (x2 - x1) / 3

        digits = [char for char in digits]

        while len(digits) < 3:

            digits.insert((1 if '-' in digits else 0), '0')

        digits = ''.join(digits)

    for d in range(len(digits)):

        draw_counter_number(x1 + (space_per_digit * d),
                            y1,
                            x1 + (space_per_digit * (d + 1)),
                            y2,
                            digits[d])
    

def draw_counter_number(x1: int, y1: int, x2: int, y2: int, n: str) -> None:
    """
    Given an area to work with, it draws a number (or a hyphen) in a
    counter.
    """

    if n not in "-0123456789":

        raise Exception(f"Parameter {n = } should be a decimal digit ('0'-'9') or a hyphen character ('-').")

    if n == '-': number = (0, 0, 0, 1, 0, 0, 0)

    elif n == '0': number = (1, 1, 1, 0, 1, 1, 1)

    elif n == '1': number = (0, 0, 1, 0, 0, 0, 1)

    elif n == '2': number = (0, 1, 1, 1, 1, 1, 0)

    elif n == '3': number = (0, 1, 1, 1, 0, 1, 1)

    elif n == '4': number = (1, 0, 1, 1, 0, 0, 1)

    elif n == '5': number = (1, 1, 0, 1, 0, 1, 1)

    elif n == '6': number = (1, 1, 0, 1, 1, 1, 1)

    elif n == '7': number = (0, 1, 1, 0, 0, 0, 1)

    elif n == '8': number = (1, 1, 1, 1, 1, 1, 1)

    elif n == '9': number = (1, 1, 1, 1, 0, 1, 1)

    extra_x, extra_y = ((x2 - x1) / 50), ((y2 - y1) / 95) # Grid is 50x95px as template

    gamelib.draw_polygon([x1 + (extra_x * 2),
                          y1 + (extra_y * 4),
                          x1 + (extra_x * 14),
                          y1 + (extra_y * 15),
                          x1 + (extra_x * 14),
                          y1 + (extra_y * 35),
                          x1 + (extra_x * 2),
                          y1 + (extra_y * 45)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[0] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 4),
                          y1 + (extra_y * 2),
                          x1 + (extra_x * 15),
                          y1 + (extra_y * 14),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 14),
                          x1 + (extra_x * 45),
                          y1 + (extra_y * 2)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[1] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 48),
                          y1 + (extra_y * 4),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 15),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 35),
                          x1 + (extra_x * 48),
                          y1 + (extra_y * 45)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[2] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 5),
                          y1 + (extra_y * 47),
                          x1 + (extra_x * 10),
                          y1 + (extra_y * 40),
                          x1 + (extra_x * 39),
                          y1 + (extra_y * 40),
                          x1 + (extra_x * 45),
                          y1 + (extra_y * 47),
                          x1 + (extra_x * 39),
                          y1 + (extra_y * 54),
                          x1 + (extra_x * 10),
                          y1 + (extra_y * 54)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[3] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 2),
                          y1 + (extra_y * 49),
                          x1 + (extra_x * 15),
                          y1 + (extra_y * 59),
                          x1 + (extra_x * 15),
                          y1 + (extra_y * 79),
                          x1 + (extra_x * 2),
                          y1 + (extra_y * 90)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[4] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 4),
                          y1 + (extra_y * 92),
                          x1 + (extra_x * 15),
                          y1 + (extra_y * 80),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 80),
                          x1 + (extra_x * 45),
                          y1 + (extra_y * 92)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[5] else '2'}"])

    gamelib.draw_polygon([x1 + (extra_x * 48),
                          y1 + (extra_y * 49),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 59),
                          x1 + (extra_x * 35),
                          y1 + (extra_y * 79),
                          x1 + (extra_x * 48),
                          y1 + (extra_y * 90)],
                          outline=PROFILE["COUNTER_OUTLINE"],
                          fill=PROFILE[f"COUNTER_{'1' if number[6] else '2'}"])



def draw_figure(x1: int, y1: int, x2: int, y2: int, grid_cols: int, grid_rows: int, colors_dict: dict) -> None:
    """
    ______________________________________________________________________

    x1, y1, x2, y2: <int>

    grid_cols, grid_rows: <int>

    colors_dict: <dict> --> {<tuple> : <str>, <tuple> : <str>, ... , <tuple> : <str>}
    --> {(<int>, <int>) : <str>, (<int>, <int>) : <str>, ... , (<int>, <int>) : <str>}


    ---> None
    ______________________________________________________________________

    Given an area to play with, the grid sizes into which the space is divided,
    and a dictionary with all colors to fill, this function draws any pixel art as
    it draws a color rectangle on any given coordinates inside the area and designated
    grid.
    """

    cell_x = (x2 - x1) / grid_cols
    cell_y = (y2 - y1) / grid_rows

    for x in range(grid_cols):

        for y in range(grid_rows):

            gamelib.draw_rectangle((x * cell_x) + x1,
                                   (y * cell_y) + y1,
                                   ((x + 1) * cell_x) + x1,
                                   ((y + 1) * cell_y) + y1,
                                   outline='', fill=(colors_dict.get((x, y), '')))

def draw_logo(game: 'MinesweeperGame') -> None:
    """
    Draws the NLGS logo and its variations.
    """

    cols, _ = game.dimensions()

    humor = (f"NLGS_{game.humor}" if (game.humor and f'NLGS_{game.humor}' in FIGURES) else "NLGS_NORMAL") + (f"_MUTE" if not game.has_audio else '')

    draw_figure(game.cell_size * (cols / 2) - (game.gui_size / 2),
                0,
                game.cell_size * (cols / 2) + (game.gui_size / 2),
                game.gui_size,
                26,
                26,
                (FIGURES[humor]))

def draw_screen(game: 'MinesweeperGame') -> None:
    """
    Draws the entirety of the elements on the screen.
    """

    draw_grid(game)
    draw_GUI(game)

    draw_buttons(game)

    draw_counter(game, 0, 0, game.width * 0.15, game.gui_size)

    draw_logo(game)


"""
Gamelib is a pure-Python single-file library/framework for writing simple games. It is
intended for educational purposes (e.g. to be used in basic programming courses).

https://github.com/dessaya/python-gamelib
"""

import tkinter as tk
from tkinter.font import Font
from tkinter import simpledialog, messagebox
from queue import Queue, Empty
from enum import Enum
import threading
import time
import signal
import os
import sys

class _TkWindow(tk.Tk):
    instance = None
    initialized = threading.Event()
    commands = Queue()

    busy_count = 0
    idle = threading.Event()
    idle.set()

    def __init__(self):
        super().__init__()

        self.closed = False

        self.title("Gamelib")
        self.resizable(False, False)

        self.assets = {}

        self.canvas = tk.Canvas(background='black')
        self.canvas.grid(column=0, row=0, sticky="nwes")

        for event_type in EventType:
            self.bind(f"<{event_type.name}>", self.handle_event)
        self.bind(f"<<notify>>", self.process_commands)
        self.protocol("WM_DELETE_WINDOW", self.close)

        self.canvas.focus_set()
        self.after_idle(self.process_commands)

    def close(self):
        self.closed = True
        self.quit()
        self.update()

    def notify(self):
        if not self.closed:
            self.event_generate('<<notify>>', when='tail')

    def process_commands(self, *args):
        _TkWindow.busy_count += 1
        _TkWindow.idle.clear()
        try:
            while True:
                try:
                    method, *args = _TkWindow.commands.get(False)
                    getattr(self, method)(*args)
                except Empty:
                    break
        finally:
            _TkWindow.busy_count -= 1
            if _TkWindow.busy_count == 0:
                _TkWindow.idle.set()

    def handle_event(self, tkevent):
        _GameThread.events.put(Event(tkevent))

    def resize(self, w, h):
        self.canvas.configure(width=w, height=h)

    def clear(self):
        self.canvas.delete("all")

    def icon(self, path):
        self.tk.call('wm', 'iconphoto', self._w, self.get_image(path))

    def draw_image(self, path, x, y):
        self.canvas.create_image(x, y, anchor='nw', image=self.get_image(path))

    def draw(self, type, args, kwargs):
        options = {'fill': 'white'}
        options.update(kwargs)
        getattr(self.canvas, f'create_{type}')(*args, **options)

    def draw_text(self, text, x, y, font, size, bold, italic, kwargs):
        options = {'fill': 'white'}
        options.update(kwargs)
        self.canvas.create_text(x, y, text=text, font=self.get_font(font, size, bold, italic), **options)

    def get_font(self, family, size, bold, italic):
        weight = 'normal'
        if bold:
            weight = 'bold'
        slant = 'roman'
        if italic:
            slant = 'italic'
        name = f'font-{family}-{size}-{weight}-{slant}'
        if name not in self.assets:
            self.assets[name] = Font(family=family, size=size, weight=weight, slant=slant)
        return self.assets[name]

    def get_image(self, path):
        if path not in self.assets:
            check_image_format(path)
            self.assets[path] = tk.PhotoImage(file=path)
        return self.assets[path]

    def say(self, message, done):
        messagebox.showinfo(self.title(), message, parent=self)
        done.put(True)

    def input(self, prompt, response):
        response.put(simpledialog.askstring(self.title(), prompt, parent=self))

    def with_window(self, func, args):
        func(self, *args)

def check_image_format(path):
    "Produce a warning message if the image format is not supported"
    ext = path[-4:].lower()
    supported = (".gif", ".ppm", ".pgm", ".pbm")
    if ext not in supported:
        print(f"{path}: Warning: image format {ext} is not supported and may not work properly on some platforms (Windows/Mac/Linux).")
        print(f"Please use one of: {supported}.")

def check_audio_format(path):
    "Produce a warning message if the audio format is not supported"
    ext = path[-4:].lower()
    if ext != ".wav":
        print(f"{path}: Warning: audio format {ext} is not supported and may not work properly on some platforms (Windows/Mac/Linux).")
        print(f"Please use WAV.")

def _audio_init():
    # shamelessly stolen from https://github.com/TaylorSMarks/playsound

    class PlaysoundException(Exception):
        pass

    def _playsoundWin(sound):
        from ctypes import c_buffer, windll
        from random import random
        from sys    import getfilesystemencoding

        def winCommand(*command):
            buf = c_buffer(255)
            command = ' '.join(command).encode(getfilesystemencoding())
            errorCode = int(windll.winmm.mciSendStringA(command, buf, 254, 0))
            if errorCode:
                errorBuffer = c_buffer(255)
                windll.winmm.mciGetErrorStringA(errorCode, errorBuffer, 254)
                exceptionMessage = ('\n    Error ' + str(errorCode) + ' for command:'
                                    '\n        ' + command.decode() +
                                    '\n    ' + errorBuffer.value.decode(getfilesystemencoding(), 'ignore'))
                raise PlaysoundException(exceptionMessage)
            return buf.value

        alias = 'playsound_' + str(random())
        winCommand('open "' + sound + '" alias', alias)
        winCommand('set', alias, 'time format milliseconds')
        durationInMS = winCommand('status', alias, 'length')
        winCommand('play', alias, 'from 0 to', durationInMS.decode())

    def _playsoundOSX(sound):
        from AppKit     import NSSound
        from Foundation import NSURL

        if '://' not in sound:
            if not sound.startswith('/'):
                sound = os.getcwd() + '/' + sound
            sound = 'file://' + sound
        url   = NSURL.URLWithString_(sound)
        nssound = NSSound.alloc().initWithContentsOfURL_byReference_(url, True)
        if not nssound:
            raise IOError('Unable to load sound named: ' + sound)
        nssound.play()

    def _playsoundNix(sound):
        from urllib.request import pathname2url

        import gi
        gi.require_version('Gst', '1.0')
        from gi.repository import Gst

        Gst.init(None)

        playbin = Gst.ElementFactory.make('playbin', 'playbin')
        if sound.startswith(('http://', 'https://')):
            playbin.props.uri = sound
        else:
            playbin.props.uri = 'file://' + pathname2url(os.path.abspath(sound))

        set_result = playbin.set_state(Gst.State.PLAYING)
        if set_result != Gst.StateChangeReturn.ASYNC:
            raise PlaysoundException(
                "playbin.set_state returned " + repr(set_result))

        bus = playbin.get_bus()
        bus.add_signal_watch()
        def on_message(bus, message):
            if message.type in (Gst.MessageType.EOS, Gst.MessageType.ERROR):
                playbin.set_state(Gst.State.NULL)
        bus.connect("message", on_message)

    from platform import system
    system = system()

    def play_sound(sound):
        """
        Play a sound located at the given path.

        Example:
            ```
            gamelib.play_sound('sound/jump.wav')
            ```

        Note:
            The only sound format that is supported accross all platforms (Windows/Mac/Linux)
            is WAV.
        """

        check_audio_format(sound)
        if system == 'Windows':
            _playsoundWin(sound)
        elif system == 'Darwin':
            _playsoundOSX(sound)
        else:
            _playsoundNix(sound)

    return play_sound

class _GameThread(threading.Thread):
    instance = None
    initialized = threading.Event()
    events = Queue()

    def start(self, game_main, args):
        self.game_main = game_main
        self.args = args
        super().start()

    def run(self):
        try:
            self.game_main(*self.args)
        except Exception as e:
            sys.excepthook(*sys.exc_info())
        finally:
            self.send_command_to_tk('close', notify=True)

    def notify_tk(self):
        self.wait_for_tk()
        w = _TkWindow.instance
        if w:
            w.notify()

    def wait_for_tk(self):
        if not _TkWindow.initialized.is_set():
            _GameThread.initialized.set()

            # block until Tk is initialized
            _TkWindow.initialized.wait()

    def send_command_to_tk(self, *args, notify=False):
        _TkWindow.commands.put(args)
        if notify:
            self.notify_tk()

    def wait(self, event_type=None):
        """
        Wait until the next `Event`: a key is pressed/released, the mouse is moved, etc,
        and return it.

        This function is normally used in combination with `gamelib.is_alive`,
        in turn-based games.

        Args:
            event_type: If an `EventType` is passed, the function will ignore any
                        events that are not of this type. (It will still return `None`
                        when the game is closed).

        Returns:
            An `Event`, or `None` if the user closed the game window.

        Example:
            ```
            while gamelib.is_alive():
                event = gamelib.wait(gamelib.EventType.KeyPress):
                gamelib.say(f'You pressed {event.key}')
            ```
        """
        self.notify_tk()
        if not _TkWindow.instance:
            return None
        while True:
            event = _GameThread.events.get()
            if not event or not event_type or event.type == event_type:
                return event

    def get_events(self):
        """
        Get the list of `Event`s that happened since the last call to `get_events`.

        This function is normally used in combination with `loop`, in action games.

        Example:
            ```
            while gamelib.loop(fps=30):
                # this is executed 30 times per second
                for event in gamelib.get_events():
                    if event.type == gamelib.EventType.KeyPress and event.key == 'q':
                        return
            ```
        """
        self.notify_tk()
        events = []
        while True:
            try:
                event = _GameThread.events.get(False)
                if not event:
                    break
                events.append(event)
            except Empty:
                break
        return events

    def title(self, s):
        """Set the window title to `s`."""
        self.send_command_to_tk('title', s)

    def icon(self, path):
        """
        Set the window icon to the image located at `path`.

        Example:
            ```
            gamelib.icon('images/icon.gif')
            ```

        Note:
            The only image formats that are supported accross all platforms (Windows/Mac/Linux)
            are GIF and PPM/PGM/PBM.
        """
        self.send_command_to_tk('icon', path)

    def draw_begin(self):
        """
        Clear the window.

        Any call to `draw_*` should be between `draw_begin` and `draw_end`.

        Example:
            ```
            gamelib.draw_begin()
            gamelib.draw_rectangle(0, 0, 10, 10, fill='red')
            gamelib.draw_end()
            ```
        """
        _TkWindow.idle.wait()
        self.send_command_to_tk('clear')

    def draw_image(self, path, x, y):
        """
        Draw an image located at `path` in the coordinates `x, y`.

        Example:
            ```
            gamelib.draw_image('images/player.gif', 10, 10)
            ```

        Note:
            The only image formats that are supported accross all platforms (Windows/Mac/Linux)
            are GIF and PPM/PGM/PBM.
        """
        self.send_command_to_tk('draw_image', path, x, y)

    def draw_text(self, text, x, y, font=None, size=12, bold=False, italic=False, **options):
        """
        Draw some `text` at coordinates `x, y` with the given properties.

        Args:
            text: The text to draw.
            x:    The screen coordinates for the text.
            y:    The screen coordinates for the text.
            font: Font family name (eg: `'Helvetica'`). **Note:** the only font guaranteed to be
                  available in all systems is the default font. If the selected font is not found,
                  the default font will be used instead.
            size: Size of the text.
            bold: Whether or not to use bold weight.
            italic: Whether or not to use italic slant.

        Some of the supported extra options are:

        * `fill`: Fill color. It can be named colors like `'red'`, `'white'`, etc,
          or a specific color in `'#rrggbb'` hexadecimal format.
        * `anchor`: Where to place the text relative to the given position.
          It may be any combination of `n` (North), `s` (South), `e`
          (East), `w` (West) and `c` (center). Default is `c`.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_text`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_text.html).

        Example:
            ```
            gamelib.draw_text('Hello world!', 10, 10, fill='red', anchor='nw')
            ```
        """
        self.send_command_to_tk('draw_text', text, x, y, font, size, bold, italic, options)

    def draw_arc(self, x1, y1, x2, y2, **options):
        """
        Draw an arc, pieslice, or chord in the bounding box between points `x1, y1` and
        `x2, y2`.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_arc`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_arc.html).

        Example:
            ```
            gamelib.draw_arc(10, 10, 20, 20, outline='white', fill='red')
            ```
        """
        self.send_command_to_tk('draw', 'arc', [x1, y1, x2, y2], options)

    def draw_line(self, x1, y1, x2, y2, **options):
        """
        Draw a straight line between points `x1, y1` and `x2, y2`.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_line`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_line.html).

        Example:
            ```
            gamelib.draw_line(10, 10, 30, 20, fill='blue', width=2)
            ```
        """
        self.send_command_to_tk('draw', 'line', [x1, y1, x2, y2], options)

    def draw_oval(self, x1, y1, x2, y2, **options):
        """
        Draw an ellipse in the bounding box between points `x1, y1` and `x2, y2`.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_oval`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_oval.html).

        Example:
            ```
            gamelib.draw_oval(10, 10, 30, 20, outline='white', fill='red')
            ```
        """
        self.send_command_to_tk('draw', 'oval', [x1, y1, x2, y2], options)

    def draw_polygon(self, points, **options):
        """
        Draw a polygon with vertices in the given `points` coordinates list. The list must have
        an even amount of numbers; each pair determines a vertex. The last vertex is automatically
        joined with the first one with a segment.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_polygon`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_polygon.html).

        Example:
            ```
            gamelib.draw_polygon([10, 10, 30, 20, 0, 40], outline='white', fill='red')
            ```
        """
        self.send_command_to_tk('draw', 'polygon', points, options)

    def draw_rectangle(self, x1, y1, x2, y2, **options):
        """
        Draw an rectangle in the bounding box between points `x1, y1` and `x2, y2`.

        To see all supported options, see the documentation for
        [`tkinter.Canvas.create_rectangle`](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/create_rectangle.html).

        Example:
            ```
            gamelib.draw_rectangle(10, 10, 30, 20, outline='white', fill='red')
            ```
        """
        self.send_command_to_tk('draw', 'rectangle', [x1, y1, x2, y2], options)

    def draw_end(self):
        """
        Refresh the window.

        Any call to `draw_*` should be between `draw_begin` and `draw_end`.

        Example:
            ```
            gamelib.draw_begin()
            gamelib.draw_rectangle(0, 0, 10, 10, fill='red')
            gamelib.draw_end()
            ```
        """
        self.send_command_to_tk('update', notify=True)

    def resize(self, w, h):
        """Resize the window to be `w` pixels wide and `h` pixels tall."""
        self.send_command_to_tk('resize', w, h)

    def say(self, message):
        """Present the user with the given `message` in a dialog box with an OK button."""
        done = Queue()
        self.send_command_to_tk('say', message, done, notify=True)
        done.get()

    def input(self, prompt):
        """
        Ask the user to enter a text value.

        Args:
            prompt: A message to display.

        Returns:
            A string containing the value that the user typed. `None` if the user
            clicked on Cancel instead of OK.
        """
        response = Queue()
        self.send_command_to_tk('input', prompt, response, notify=True)
        return response.get()

    def is_alive(self):
        """
        Returns True if the game window is open.

        Example:
            ```
            while gamelib.is_alive():
                event = gamelib.wait(gamelib.EventType.KeyPress):
                gamelib.say(f'You pressed {event.key}')
            ```
        """
        self.wait_for_tk()
        return bool(_TkWindow.instance)

    _last_loop_time = None

    def loop(self, fps=30):
        """
        When used in a `while` loop, the body will be executed `fps` times per second.

        Returns:
            `True` if the game window is still open, `False` otherwise.

        Example:
            ```
            while gamelib.loop(fps=30):
                # this is executed 30 times per second
                for event in gamelib.get_events():
                    if event.type == gamelib.EventType.KeyPress and event.key == 'q':
                        return
            ```
        """
        frame_duration = 1.0 / fps
        a = _GameThread._last_loop_time
        b = time.time()
        if a:
            time.sleep(max(0, frame_duration - (b - a)))
        _GameThread._last_loop_time = time.time()
        return self.is_alive()

_GameThread.instance = _GameThread()

wait = _GameThread.instance.wait
get_events = _GameThread.instance.get_events
title = _GameThread.instance.title
icon = _GameThread.instance.icon
draw_begin = _GameThread.instance.draw_begin
draw_image = _GameThread.instance.draw_image
draw_text = _GameThread.instance.draw_text
draw_arc = _GameThread.instance.draw_arc
draw_line = _GameThread.instance.draw_line
draw_oval = _GameThread.instance.draw_oval
draw_polygon = _GameThread.instance.draw_polygon
draw_rectangle = _GameThread.instance.draw_rectangle
draw_end = _GameThread.instance.draw_end
resize = _GameThread.instance.resize
say = _GameThread.instance.say
input = _GameThread.instance.input
is_alive = _GameThread.instance.is_alive
loop = _GameThread.instance.loop
play_sound = _audio_init()

def _sigint_handler(sig, frame):
    w = _TkWindow.instance
    if w:
        w.close()
    else:
        raise KeyboardInterrupt()

def init(game_main, args=None):
    """
    Initialize gamelib.

    Args:
        game_main: Your `main` function.
        args: List of arguments to be passed to the `main` function, or `None`.
    """
    _GameThread.instance.start(game_main, args or [])

    # block until wait(), get_events(), etc called on game thread.
    # This prevents rendering the window before the user has a chance to configure it.
    _GameThread.initialized.wait()

    _TkWindow.instance = _TkWindow()
    _TkWindow.initialized.set()

    signal.signal(signal.SIGINT, _sigint_handler)

    try:
        _TkWindow.instance.mainloop()
    finally:
        _GameThread.events.put(None)
        _TkWindow.instance = None
        _GameThread.instance.join(1)
        if _GameThread.instance.is_alive():
            print('Killing unresponsive game thread. Make sure to call get_events() or wait() periodically.')
            os._exit(1)
        os._exit(0)

class EventType(Enum):
    "An enumeration of the different types of `Event`s supported by gamelib."

    KeyPress = 'KeyPress'
    "The user pressed a key."
    KeyRelease = 'KeyRelease'
    "The user released a key."
    Motion = 'Motion'
    "The user moved the mouse over the window."
    ButtonPress = 'ButtonPress'
    "The user pressed a mouse button."
    ButtonRelease = 'ButtonRelease'
    "The user released a mouse button."

class Event:
    """
    Represents an event generated by the user.

    Attributes:
        type: An `EventType`.
        key: A key that has been pressed/released.
        mouse_button: 0, 1 or 2 for left, right and middle mouse buttons respectively.
        x: The current mouse horizontal position, in pixels.
        y: The current mouse vertical position, in pixels.

    This is actually a wrapper for the
    [Tkinter Event class](https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/event-handlers.html).
    Any of the `tk.Event` attributes can be accessed through this object.

    ## See also

    `wait`, `get_events`
    """

    def __init__(self, tkevent):
        self.tkevent = tkevent

    def __getattr__(self, k):
        if k == 'type': return EventType[self.tkevent.type.name]
        if k == 'key': return self.tkevent.keysym
        if k == 'mouse_button': return self.tkevent.num
        return getattr(self.tkevent, k)

    def __repr__(self):
        return repr(self.tkevent)

if __name__ == '__main__':
    def interactive_main(_locals):
        import code
        code.interact(local=_locals)

    init(interactive_main, args=[locals()])



"""
This module manages all text files, and stores function related
to reading or writing in them.
"""

from .minepaths import MINE_FIGURES, MINE_PROFILES

def map_color_profiles(file_name: str=MINE_PROFILES) -> dict:
    """
    ______________________________________________________________________

    file_name: <str>


    ---> <dict> --> {<str> : <dict>*, <str> : <dict>*, ... , <str> : <dict>*}

        *<dict> --> {<str> : <str>, <str> : <str>, ... , <str> : <str>}
    ______________________________________________________________________

    Returns a dictionary with other dictionaries as its values which stores the
    color values for each personalized color profile.
    """

    profiles_dict = dict()
    current_name = ''

    with open(file_name) as file:

        for line in file:

            if line == '\n' or line[0] == '#': # line defines a comment

                continue

            type, key, *value = ''.join(line.split('=')).split()

            if type == "!t": # line defines a new Profile

                profiles_dict[key] = dict()
                current_name = key

            elif type == "!v": # line defines a 'key-value' pair

                value = ''.join(value)
                profiles_dict[current_name][key] = (value if not value == '/' else '')

    return profiles_dict

def map_figures(file_name: str=MINE_FIGURES) -> dict:
    """
    ______________________________________________________________________

    file_name: <str>


    ---> <dict> --> {<str> : <dict>*, <str> : <dict>*, ... , <str> : <dict>*}

        *<dict> --> {<tuple>** : <str>, <tuple>** : <str>, ... , <tuple>** : <str>}

        **<tuple> --> (<int>, <int>)
    ______________________________________________________________________

    Returns a dictionary with other dictionaries as its values which stores the
    color values for each pixel in a defined area of a figure.
    """

    figures_dict = dict()
    current_figure = ''

    figure_list = list()
    searching_in_list = False

    with open(file_name) as file:

        for line in file:

            if line == '\n' or line[0] == '#': # line is empty or defines a comment

                continue

            fragmented_line = line.rstrip().split()

            if searching_in_list and not fragmented_line[0] == "/>":

                figure_list.append(fragmented_line)

            if fragmented_line[0] == "!f":

                name = fragmented_line[1]

                figures_dict[name] = dict()
                current_figure = name

            elif fragmented_line[0] == "!<":

                figure_list = list()
                searching_in_list = True

            elif fragmented_line[0] == "/>":

                for px_x in range(len(figure_list[0])):

                    for px_y in range(len(figure_list)):

                        pixel = figure_list[px_y][px_x]

                        if not pixel == '.':

                            figures_dict[current_figure][(px_x, px_y)] = pixel

                searching_in_list = False

    return figures_dict


"""
This is a simplified Minesweeper Game, by Franco 'NLGS' Lighterman.

'Gamelib' by Diego Essaya.

More at the GitHub Repository: https://github.com/NLGS2907/gamelib-games/tree/master/minesweeper
"""

"""
This module holds the game object and all of its logic.
"""

from random import randint
from ..menu import Button
from typing import Union
from .. import gamelib
from . import minegraphics
from .minepaths import MINE_BOOM, MINE_BTN_DN, MINE_BTN_UP, MINE_CELL, MINE_ICON, MINE_MARKED, MINE_NEW_GAME, MINE_POPUP, MINE_VICTORY

__version__ = "1.0.0"

SIZE_CONSTANT = 30

MINE = 'X'

INSTRUCTIONS = """\t-=-=\tINSTRUCTIONS\t=-=-


How To Play:

    Your job is to uncover each cell of the grid without
    stepping on a mine. If that happens, you loose the
    game, as you only have one life.

    Each cell that is not a mine has a number assigned
    to it. This number indicates the numbers of mines
    that are adjacent to this cell, either vertically,
    horizontally, or diagonally. Use this information
    to your advantage.

    With a bit of logic (and some guessing) you will
    surely persevere. Good Luck!

Grid:

    * [Left Click] to reveal a cell.

    * [Right Click] to cycle through unmarked, flag, or 'dunno'.

    * [Middle Click] to reveal al adjacent cells,
      if neighbours are flagged enough.

GUI:

    * [Left Click] on 'NLGS Logo' to start new game.

    * [Right Click] on 'NLGS Logo' to exit the game.

    * [Middle Click] on 'NLGS Logo' to mute or unmute
      the game

    Click the buttons of 'Mines', 'X' and 'Y' to select
    the parameters of the new game.

    - 'X' is the number of COLUMNS of the new grid.

    - 'Y' is the number of ROWS of the new grid.

    - 'Mines' is the amount of mines to be used in the
      new grid.
      If it is set to 'Auto' it will automatically use 20% of
      the new grid.
"""

def is_in_bounds(x: int, y: int, max_x: int, max_y: int) -> bool:
    """
    Given the max values for the width and height of an area (minimum value is
    always 0 for both), it calculates if the coordinates '(x, y)' are within
    said area.
    """

    return (0 <= x <= max_x) and (0 <= y <= max_y)

class MinesweeperCell:
    """
    Tiny class for a cell representation.
    """

    def __init__(self, value: Union[str, int]) -> None:
        """
        Initializes an instance of type 'MinesweeperCell'.
        """

        self._is_hidden = True
        self.value = value
        """
        The value of MINE, or a value between 0 and 8 inclusive.

        Can also be an empty string, to indicate that an empty cell already has been visited.
        """
        self.state = 0
        """
        '0' for an unmarked cell.

        '1' for a flagged cell.

        '2' for a cell with the symbol '?'.

        All the states can cycle over.
        """
    
    @property
    def is_hidden(self) -> bool:
        """
        Defines the property 'is_hidden' of the cell.
        """
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, new_value: bool) -> None:
        """
        Changes the value of 'self.is_hidden' so the cell is now uncovered.

        One cell should never be covered again. As such, no value other than
        'False' is accepted.
        """

        if new_value == False:

            self._is_hidden = new_value

    def cycle_state(self) -> None:
        """
        Cycles through the possible stages of a cell, if it is uncovered.
        """
        
        if self.is_hidden:

            self.state = (self.state + 1) % 3

class MinesweeperGame:
    """
    Class for holding variables of a Minesweeper game.
    """

    def __init__(self, size_constant: int, grid_cols: int=20, grid_rows: int=20, mines: int=0) -> None:
        """
        Initializes an instance of type 'MinesweeperGame'.

        'size_constant' is the size of each cell (individual width
        and heght are the same).

        'grid_cols' means the columns of the game grid.

        'grid_rows' means the rows of the game grid.

        'mines' is the amount of mines in the game grid.
        """

        if grid_cols < 10 or grid_rows < 10:

            raise Exception(f"Invalid grid dimensions: Size given is ({grid_cols}, {grid_rows}). Minimum accepted is (10, 10)")

        if mines and mines > (grid_cols * grid_rows):

            mines = grid_cols * grid_rows

        # Sizes
        self.grid_x, self.grid_y = grid_cols, grid_rows

        self.cell_size = size_constant

        self.width = self.cell_size * self.grid_x
        self.gui_size = self.width // 12
        self.height = (self.cell_size * self.grid_y) + self.gui_size

        self.mines = mines
        mines_to_use = (mines if mines > 0 else int(self.grid_x * self.grid_y * 0.2)) # If mines is note specified, use 20% of the grid.
        self.flags = mines_to_use

        self.grid = self.generate_grid(self.grid_x, self.grid_y, mines_to_use)

        # When the player loses, this cell will be highlighted
        self.loser_cell = None

        # Buttons
        self.buttons = self.generate_buttons()

        # NLGS Humor
        self.humor = "NORMAL"

        # Control Booleans
        self.has_audio = True
        self.end = False
        self.exit = False

    def dimensions(self) -> tuple[int, int]:
        """
        Returns the dimensions of the game grid.
        """

        return len(self.grid[0]), len(self.grid)

    def generate_grid(self, how_many_columns: int, how_many_rows: int, how_many_mines: int) -> list:
        """
        Generates the game grid.
        """

        grid = list()
        mines_set = set()

        while len(mines_set) < how_many_mines:

            x, y = randint(0, (how_many_columns - 1)), randint(0, (how_many_rows - 1))

            mines_set.add((x, y))

        for row in range(how_many_rows):

            row_list = list()

            for col in range(how_many_columns):
                
                if (col, row) in mines_set:

                   row_list.append(MinesweeperCell(MINE))
                   mines_set.remove((col, row)) # Don't need it anymore

                else:

                    row_list.append(MinesweeperCell(0)) # This generates [0, 0], but the first 0 is not the same as HIDDEN 

            grid.append(row_list)

        return self.generate_numbers(grid)

    def generate_numbers(self, grid: list) -> list:
        """
        Generate the numbers which tell how many mines is around them.
        """

        cols, rows = len(grid[0]), len(grid)

        for y in range(rows):

            for x in range(cols):

                if grid[y][x].value == MINE: # Ignore if it is a mine

                    continue

                count = 0

                for dx in range(-1, 2):

                    for dy in range(-1, 2):

                        if is_in_bounds((x + dx), (y + dy), (cols - 1), (rows - 1)) and grid[y + dy][x + dx].value == MINE:

                            count += 1

                grid[y][x].value = count

        return grid

    def generate_buttons(self) -> list:
        """
        ______________________________________________________________________

        ---> <list> --> [<Button>, <Button>, ... , <Button>]
        ______________________________________________________________________

        Generates the buttons of the game.
        """

        buttons_list = list()
        self.new_game = Button(self.width * 0.450333333,
                               0,
                               self.width * 0.549666666,
                               self.gui_size)

        buttons_list.append(self.new_game)

        self.cols_dn = Button(self.width * 0.6,
                              0,
                              self.width * 0.65,
                              self.gui_size,
                              '<')

        buttons_list.append(self.cols_dn)

        self.cols_up = Button(self.width * 0.75,
                              0,
                              self.width * 0.8,
                              self.gui_size,
                              '>')

        buttons_list.append(self.cols_up)

        self.rows_dn = Button(self.width * 0.8,
                              0,
                              self.width * 0.85,
                              self.gui_size,
                              '<')

        buttons_list.append(self.rows_dn)

        self.rows_up = Button(self.width * 0.95,
                              0,
                              self.width,
                              self.gui_size,
                              '>')

        buttons_list.append(self.rows_up)

        self.mines_dn = Button(self.width * 0.2,
                              0,
                              self.width * 0.25,
                              self.gui_size,
                              '<')

        buttons_list.append(self.mines_dn)

        self.mines_up = Button(self.width * 0.35,
                              0,
                              self.width * 0.4,
                              self.gui_size,
                              '>')

        buttons_list.append(self.mines_up)

        self.info = Button(self.width * 0.15,
                              0,
                              self.width * 0.2,
                              self.gui_size,
                              'i')

        buttons_list.append(self.info)

        return buttons_list

    def search_coords(self, x: int, y: int) -> tuple[int, int]:
        """
        Given the (x, y) coordinates (in pixels) of a click, it
        searches to which cell of the grid it corresponds to.
        """

        cell_x = x // self.cell_size
        cell_y = (y - self.gui_size) // self.cell_size

        return cell_x, cell_y

    def reveal_mines(self) -> None:
        """
        Uncovers all the cells that contain a main.

        This function is meant to be called once and only once
        the game ended.
        """

        cols, rows = self.dimensions()

        for x in range(cols):

            for y in range(rows):

                cell = self.grid[y][x]

                if cell.value == MINE and not cell.state == 1: # Is a mine and is not flagged

                    cell.is_hidden = False

    def reveal_cells(self, x: int, y: int) -> None:
        """
        Uncovers the content of a cell and also of its neighbours,
        if the value is 0.
        """

        cell = self.grid[y][x]
        cell.is_hidden = False

        cols, rows = self.dimensions()

        for dx in range(-1, 2):

            for dy in range(-1, 2):

                if is_in_bounds((x + dx), (y + dy), (cols - 1), (rows - 1)):

                    neighbour = self.grid[y + dy][x + dx]
                    if neighbour.state == 0: neighbour.is_hidden = False

                    if neighbour.value == 0 and neighbour.state == 0:

                        neighbour.value = ''
                        self.reveal_cells(x + dx, y + dy)


    def process_click(self, x: int, y: int, click_type: int) -> None:
        """
        Given the coordinates and the type of the click, it processes
        the click action into its corresponding actions.

        Returns a boolean that says if an action was performed with
        such click.
        """

        if 0 <= x <= self.width and self.gui_size < y <= self.height and not self.end:

            cell_x, cell_y = self.search_coords(x, y)

            cell = self.grid[cell_y][cell_x]

            if not cell.is_hidden: # Cell was clicked already.

                if click_type == 2: # Is midddle click.

                    flag_count = 0

                    for dx in range(-1, 2):

                        for dy in range(-1, 2):

                            cols, rows = self.dimensions()

                            if is_in_bounds((cell_x + dx), (cell_y + dy), (cols - 1), (rows - 1)):

                                neighbour = self.grid[cell_y + dy][cell_x + dx]

                                if neighbour.state == 1: # If it is flagged.

                                    flag_count += 1

                    if flag_count >= (cell.value if not cell.value == '' else 0):

                        for dx in range(-1, 2):

                            for dy in range(-1, 2):

                                if is_in_bounds((cell_x + dx), (cell_y + dy), (cols - 1), (rows - 1)):

                                    neighbour = self.grid[cell_y + dy][cell_x + dx]

                                    if neighbour.state == 0:
                                        
                                        neighbour.is_hidden = False

                                        if neighbour.value == MINE:

                                            self.loser_cell = neighbour
                                            self.reveal_mines()

                                        elif neighbour.value == 0:

                                            self.reveal_cells((cell_x + dx), (cell_y + dy))

            if click_type == 3: # Is secondary click.

                cell.cycle_state()

                if cell.state == 1: # Is flagged

                    self.flags -= 1

                elif cell.state == 2:

                    self.flags += 1

                if cell.is_hidden and self.has_audio: gamelib.play_sound(MINE_MARKED)

            elif click_type == 1 and cell.state == 0: # Is primary click and cell is unmarked.

                if cell.value == MINE:

                    self.loser_cell = cell
                    self.reveal_mines()

                elif cell.value == 0:

                    self.reveal_cells(cell_x, cell_y)

                if cell.is_hidden and self.has_audio:
                    
                    gamelib.play_sound(MINE_CELL.format(cell=cell.value))

                cell.is_hidden = False

        else:

            for button in self.buttons:

                if button.is_inside(x, y):

                    if button is self.new_game:

                        if click_type == 1:

                            self.init_new_game(self.grid_x, self.grid_y, self.mines)

                        elif click_type == 2:

                            self.has_audio = not self.has_audio

                        elif click_type == 3:

                            self.exit = True

                    elif button is self.cols_up:

                        self.grid_x += 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_UP)

                    elif button is self.cols_dn:

                        if self.grid_x > 10:

                            self.grid_x -= 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_DN)

                    elif button is self.rows_up:

                        self.grid_y += 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_UP)

                    elif button is self.rows_dn:

                        if self.grid_y > 10:

                            self.grid_y -= 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_DN)

                    elif button is self.mines_up:

                        cols, rows = self.dimensions()

                        if self.mines <= 0:

                            self.mines = 1

                        elif self.mines < (cols * rows): self.mines += 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_UP)

                    elif button is self.mines_dn:

                        if self.mines and self.mines > 0:

                            self.mines -= 1

                        if self.has_audio: gamelib.play_sound(MINE_BTN_DN)

                    elif button is self.info:

                        if self.has_audio: gamelib.play_sound(MINE_POPUP)

                        gamelib.say(INSTRUCTIONS)

    def did_it_end(self) -> tuple[bool, bool]:
        """
        Returns a tuple in which the first element indicates if the
        game ended already. The second one indicates if it was a
        victory or not.
        """

        if self.loser_cell: # Landed on a mine

            if not self.end and self.has_audio: gamelib.play_sound(MINE_BOOM)

            self.humor = "DEAD"
            return True, False

        cols, rows = self.dimensions()

        for x in range(cols):

            for y in range(rows):

                cell = self.grid[y][x]

                if cell.value == MINE and cell.state == 1:

                    continue

                if cell.is_hidden:

                    return False, False

        self.humor = "HAPPY"
        if not self.end and self.has_audio: gamelib.play_sound(MINE_VICTORY)
        return True, True

    def init_new_game(self, grid_x: int, grid_y: int, mines: int) -> None:
        """
        Begins a new game with the current values for the
        columns and rows of the game grid, and also the amount
        of mines.
        """

        self.end = False
        self.loser_cell = None

        mines_to_use = (mines if mines > 0 else int(self.grid_x * self.grid_y * 0.2))
        self.flags = mines_to_use

        self.width = self.cell_size * grid_x
        self.gui_size = self.width // 12
        self.height = (self.cell_size * grid_y) + self.gui_size

        self.humor = "NORMAL"
        self.grid = self.generate_grid(grid_x, grid_y, mines_to_use)
        self.buttons = self.generate_buttons()

        gamelib.resize(self.width, self.height)

        if self.has_audio: gamelib.play_sound(MINE_NEW_GAME)


def mine_main() -> None:
    """
    Main function. Shouldn't be called more than once.
    """

    game = MinesweeperGame(SIZE_CONSTANT, 20, 20)

    gamelib.title("Plain Minesweeper! by NLGS")
    gamelib.resize(game.width, game.height)
    gamelib.icon(MINE_ICON)

    while gamelib.loop(fps=60):

        if game.exit:
            break

        gamelib.draw_begin()
        minegraphics.draw_screen(game)
        gamelib.draw_end()

        click = gamelib.wait(gamelib.EventType.ButtonPress)

        if click: game.process_click(click.x, click.y, click.mouse_button)

        ended, _ = game.did_it_end()

        if ended:
            game.end = True


if __name__ == "__main__":
    gamelib.init(mine_main)


"""
This module manages all text files, and stores function related
to reading or writing in them.
"""

from .pipaths import PICROSS_PROFILES


def map_color_profiles(file_name: str=PICROSS_PROFILES) -> dict:
    """
    ______________________________________________________________________

    file_name: <str>


    ---> <dict> --> {<str> : <dict>*, <str> : <dict>*, ... , <str> : <dict>*}

        *<dict> --> {<str> : <str>, <str> : <str>, ... , <str> : <str>}
    ______________________________________________________________________

    Returns a dictionary with other dictionaries as its values which stores the
    color values for each personalized color profile.
    """

    profiles_dict = dict()
    current_name = ''

    with open(file_name) as file:

        for line in file:

            if line == '\n' or line[0] == '#': # line defines a comment

                continue

            type, key, *value = ''.join(line.split('=')).split()

            if type == "!t": # line defines a new Profile

                profiles_dict[key] = dict()
                current_name = key

            elif type == "!v": # line defines a 'key-value' pair

                value = ''.join(value)
                profiles_dict[current_name][key] = (value if not value == '/' else '')

    return profiles_dict



from .gamelib import init as game_init
from .picross import pi_main as picross_main
from .minesweeper import mine_main as minesweeper_main

__version__ = "1.1.1"

ACCEPTED_VALUES = 'picross', 'minesweeper'

def main(argv) -> None:

    if not argv or argv[0] not in ACCEPTED_VALUES:

        print(f"Game Name Invalid. Please use one of the following:\n{ACCEPTED_VALUES}")
        return

    game_chosen: str = argv[0]

    if game_chosen == 'picross':

        game_init(picross_main)

    elif game_chosen == 'minesweeper':

        game_init(minesweeper_main)


if __name__ == '__main__':

    main()



"""
Setup manifest.
"""

from setuptools import setup, find_packages

with open("README.md", mode='r', encoding="utf-8") as readme:

    project_description = readme.read()

setup(
  name = 'gamelibgames',
  packages = find_packages(),
  package_data={
    'gamelibgames': [
      "picross/img/gameplay_demo/*.gif",
      "picross/img/icon/*.gif",
      "picross/sfx/*.wav",
      "picross/*.txt",
      "picross/img/demo/*.gif",
      "picross/img/icon/*.gif",
      "picross/img/references/*.png",
      "picross/sfx/*.wav",
      "minesweeper/*.txt"
    ]
  },
  version = '1.1.3',
  license='MIT',
  description = 'A collection of minigames made with Gamelib.',
  long_description=project_description,
  long_description_content_type="text/markdown",
  author = 'NLGS',
  author_email = 'flighterman@fi.uba.ar',
  url = 'https://github.com/NLGS2907/gamelib_games',
  keywords = ['GAMELIB', 'MINIGAMES'],
  classifiers=[
    'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.10',
  ],
)



"""
This is a simplified Picross Game, by Franco 'NLGS' Lighterman.

'Gamelib' by Diego Essaya.

More at the GitHub Repository: https://github.com/NLGS2907/gamelib-games/tree/master/picross
"""

"""
This module contains the Game Object and all functions and methods that
belong to logic processing.
"""

from .. import gamelib
from ..menu import Button
from . import pigraphics
from .pipaths import PI_GAME_LOST, PI_GAME_WON, PI_TILE_WRONG, PI_TILE_RIGHT, PICROSS_ICON

from random import choice
from typing import Union, Callable

__version__ = "1.1.0"

WIDTH = 500
HEIGHT = 700

def column_list(grid: list, x: int, with_index: bool=True) -> list:
    """
    ______________________________________________________________________

    grid: <list> --> [None, <list>*, <list>*, ... , <list>*]

            *<list> --> [[<list>, <int>, <int>, ... , <int>],
                        [<list>, <int>, <int>, ... , <int>]
                        [<list>, <int>, <int>, ... , <int>]]

    x: <int>

    with_index: <bool>


    ---> <list> --> [<list>*, <int>, <int>, ... , <int>]

        *<list> --> [<int>, <int>, ... , <int>]
    ______________________________________________________________________

    Returns a list with the values of a selected column of a given grid 'grid'.

    'x' must be a valid value between 0 and the number of columns of the grid.

    'with_index' indicates if the returned list includes the index of said column.
    """

    cols, rows = len(grid[0]), len(grid)

    if x < 0 or x >= cols:

        raise Exception(f"Column {x} is out of range")

    temp_list = list()

    for i in range((0 if with_index else 1), rows):

        temp_list.append(grid[i][x])
    
    return temp_list

def row_list(grid: list, y: int, with_index: bool=True) -> list:
    """
    ______________________________________________________________________

    grid: <list> --> [None, <list>*, <list>*, ... , <list>*]

            *<list> --> [[<list>, <int>, <int>, ... , <int>],
                        [<list>, <int>, <int>, ... , <int>]
                        [<list>, <int>, <int>, ... , <int>]]

    y: <int>

    with_index: <bool>


    ---> <list> --> [<list>*, <int>, <int>, ... , <int>]

        *<list> --> [<int>, <int>, ... , <int>]
    ______________________________________________________________________

    Returns a list with the values of a selected row of a given grid 'grid'.

    'y' must be a valid value between 0 and the number of rows of the grid.

    'with_index' indicates if the returned list includes the index of said row.
    """

    cols, rows = len(grid[0]), len(grid)

    if y < 0 or y >= rows:

        raise Exception(f"Row {y} is out of range")

    temp_list = list()

    for j in range((0 if with_index else 1), cols):

        temp_list.append(grid[y][j])
    
    return temp_list

def split_filter(list_to_filter: list, filter_function: Union[Callable, None]=None) -> list:
    """
    Given a list and a filter function, it returns another list of sublists
    which each have the elements of 'list_to_filter' to which the filter function
    'filter_function' returns 'True', using those the 'filter_function' returns
    'False' from as a separator.

    example: converts from [2, 4, 5, 6, 7, 8, 10, 12] to [[2, 4], [6], [8, 10, 12]],
    should the filter function be, say, 'lambda x: x % 2 == 0' (the element is even)

    If the filter function is None, then evaluates if the elements themselves are True
    or False.
    """
    big_list = list()
    sublist = list()

    can_add = False

    for element in list_to_filter:

        if (filter_function(element) if filter_function else element):

            if element is list_to_filter[len(list_to_filter) - 1]:

                can_add = True

            sublist.append(element)
        
        else:

            can_add = True

        if can_add:

            can_add = False

            if sublist:

                big_list.append(sublist)
                sublist = list()

    return big_list

class PicrossGame:
    """
    Main class with the purpose of storing all the games variables.
    """

    def __init__(self, initial_width: int, initial_height: int, grid_cols: int=10, grid_rows: int=10, how_many_mistakes: int=3, must_discover_grid: bool=False) -> None:
        """
        Initalizes an instance of type 'PicrossGame'.

        'initial_width' and 'initial_height' are the sizes of the game window.

        'grid_cols' and 'grid_rows' are the dimensions of the game grid. The minum allowed
        is '4' for both values.

        'how_many_mistakes' is the max number of mistakes the player is allowed
        to do before the game is considered lost.

        'must_discover_grid' is an option which indicates if the games ends when the
        player discovers all marked tiles, or when they reveal all the grid, regardless
        of the cell is marked or not. It is 'False' by default.
        """

        if initial_width > initial_height:

            print(f"[WARNING]: Height should always be greater than Width...\nReversing the values from ({initial_width}, {initial_height}) tp ({initial_height}, {initial_width})")

            initial_width, initial_height = initial_height, initial_width

        if grid_cols < 4 or grid_rows < 4:

            raise Exception(f"Invalid grid dimensions: Size given is ({grid_cols}, {grid_rows}). Minimum accepted is (4, 4)")

        # Dimensions
        self.width, self.height = initial_width, initial_height

        # Even though there is a function that returns the dimensions, these can change with time,
        # and differ from the actual grid dimensions values.
        self.grid_x, self.grid_y = grid_cols, grid_rows

        self.gui_size = self.height // 7

        # Buttons
        self.buttons_list = list()
        self.generate_buttons()

        # Grid
        self.grid = self.generate_level(self.grid_x, self. grid_y)

        # Mistakes Control
        self.max_mistakes = how_many_mistakes
        self.mistakes_count = 0
        self.mistakes_coord = list()

        # Control booleans
        self.ended = False
        self.has_audio = True
        self.has_to_clear_grid = must_discover_grid

    def generate_buttons(self) -> None:
        """
        Generates the buttons that are meant to be used in the game.
        """

        cols_x, cols_y = (self.width * 0.65), (self.gui_size * 0.65)
        rows_x, rows_y = (self.width * 0.73), (self.gui_size * 0.65)

        frame_margin = int(self.gui_size / 6.666)

        self.new_game = Button((self.width / 50), (self.width / 10), (self.width / 5), (self.gui_size * 0.9), "New Game")
        self.buttons_list.append(self.new_game)

        self.cols_up = Button(cols_x - frame_margin,
                              cols_y - (2 * frame_margin),
                              cols_x + frame_margin,
                              cols_y - frame_margin,
                              "/\\")
        self.buttons_list.append(self.cols_up)

        self.cols_dn = Button(cols_x - frame_margin,
                              cols_y + frame_margin,
                              cols_x + frame_margin,
                              cols_y +  (2 * frame_margin),
                              "\/")
        self.buttons_list.append(self.cols_dn)

        self.rows_up = Button(rows_x - frame_margin,
                              rows_y - (2 * frame_margin),
                              rows_x + frame_margin,
                              rows_y - frame_margin,
                              "/\\")
        self.buttons_list.append(self.rows_up)

        self.rows_dn = Button(rows_x - frame_margin,
                              rows_y + frame_margin,
                              rows_x + frame_margin,
                              rows_y +  (2 * frame_margin),
                              "\/")
        self.buttons_list.append(self.rows_dn)

        self.audio = Button((self.width * 0.9),
                      0,
                      self.width,
                      (self.gui_size * 0.4),
                      '')
        self.buttons_list.append(self.audio)

    def generate_level(self, how_many_columns: int, how_many_rows: int) -> list:
        
        """
        ______________________________________________________________________

        how_many_columns, how_many_rows: <int>


        ---> <list> --> [<list>, <list>, ... , <list>]
        ______________________________________________________________________

        Return a list of lists which is the grid of the grid with its indexes.
        """

        grid = [[None] + [list() for _ in range(how_many_columns)]]

        for _ in range(how_many_rows):

            temp_list = [list()]

            for _ in range(how_many_columns):

                temp_list.append([choice((False, True)), True])
            
            grid.append(temp_list)

        return self.generate_index(grid)

    def generate_index(self, grid: list) -> list:
        """
        ______________________________________________________________________

        grid: <list> --> [<list>, <list>, ... , <list>]


        ---> <list> --> [<list>, <list>, ... , <list>]
        ______________________________________________________________________

        Receives a list of lists which is the grid of the game, processes it and
        writes the indexes according to its columns and rows.
        """

        cols, rows = len(grid[0]), len(grid)

        for col_index in range(1, cols):

            column = column_list(grid, col_index, with_index=False)
            filtered_column = [len(elem) for elem in split_filter(column, (lambda elem: elem[0]))]

            grid[0][col_index] = (filtered_column if filtered_column else [0])

        for row_index in range(1, rows):

            row = row_list(grid, row_index, with_index=False)
            filtered_row = [len(elem) for elem in split_filter(row, (lambda elem: elem[0]))]

            grid[row_index][0] = (filtered_row if filtered_row else [0])

        return grid

    def dimensions(self) -> list[int, int]:
        """
        Returns the dimensions of the game grid (NOT those of the window).
        """

        return (len(self.grid[0]), len(self.grid))

    def calculate_indexes(self) -> tuple[int, int]:
        """
        Returns a tuple which is the calculated spaces dedicated to the index cells.
        """

        return (self.width / 5), (self.height / 7)

    def calculate_playable_area(self) -> tuple[int, int, int, int]:
        """
        Returns a tuple which is the cornes of the area dedicated to the clickable cells.
        """

        index_x, index_y = self.calculate_indexes()

        return (index_x, (index_y + self.gui_size), self.width, self.height)

    def calculate_coords(self, x: int, y: int) -> tuple[int, int]:
        """
        Given a pair of coordinates, it traces it back to the grid and calculates whose
        cell it belongs to.
        
        Returns a tuple of two integers.
        """

        cols, rows = self.dimensions()
        index_x, index_y = self.calculate_indexes()

        cell_x = (self.width - index_x) / (cols - 1)
        cell_y = (self.height - index_y - self.gui_size) / (rows - 1)

        return int((x - index_x) / cell_x) + 1, int((y - index_y - self.gui_size) / cell_y) + 1

    def process_click(self, x: int, y: int, event_type: int) -> None:
        """
        Given the coordinates of a click action and the type of button pressed, it
        gets processed into its corresponding action.

        'event_type' is meant to be always 1, 2 or 3, but only '1 or else' is processed.
        """

        area_x1, area_y1, area_x2, area_y2 = self.calculate_playable_area()

        if area_x1 <= x <= area_x2 and area_y1 <= y <= area_y2:

            x_coord, y_coord = self.calculate_coords(x, y)

            cell = self.grid[y_coord][x_coord]

            if cell[1]:

                if (not cell[0] if event_type == 1 else cell[0]):

                    self.mistakes_count += 1
                    self.mistakes_coord.append((x_coord, y_coord))
                    if self.has_audio: gamelib.play_sound(PI_TILE_WRONG)
                
                else:

                    if self.has_audio: gamelib.play_sound(PI_TILE_RIGHT)

                cell[1] = False

        else:

            for button in self.buttons_list:

                if button.is_inside(x, y):

                    if button is self.new_game:

                        self.start_new_game()

                    elif button is self.cols_up:

                        self.grid_x += 1

                    elif button is self.cols_dn:

                        self.grid_x = ((self.grid_x - 1) if self.grid_x > 4 else self.grid_x)

                    elif button is self.rows_up:

                        self.grid_y += 1

                    elif button is self.rows_dn:

                        self.grid_y = ((self.grid_y - 1) if self.grid_y > 4 else self.grid_y)

                    elif button is self.audio:

                        self.has_audio = not self.has_audio

                    break

    def did_it_end(self) -> tuple[bool, bool]:
        """
        It determines if the game has ended, and if so, if the player won or not.

        The first boolean is to determine if the game ended or not.

        The second boolean is to determine if it was a victory ('False' by default,
        if the game hasn't ended yet).
        """

        cols, rows = self.dimensions()

        if self.mistakes_count >= self.max_mistakes:

            return True, False

        for x in range(1, cols):

            for y in range(1, rows):

                is_filled, is_hidden = self.grid[y][x]

                if (True if self.has_to_clear_grid else is_filled) and is_hidden:

                    return False, False
        
        return True, True

    def advance_game(self) -> None:
        """
        Avances the state of the game, updating the corresponding attributes if
        necessary.
        """

        game_ended, is_victory = self.did_it_end()

        if game_ended:

            self.ended = True

            if self.has_audio: gamelib.play_sound(PI_GAME_WON if is_victory else PI_GAME_LOST)
            gamelib.say("You Won!" if is_victory else "You Lost!")

            answer = gamelib.input("Another Round? (y/n)")

            if answer == 'y':

                self.start_new_game()

    def start_new_game(self) -> None:
        """
        Modifies certain game attributes so a new grid is generated with the
        given modifiable values, and also resets the mistakes count, the coordinates
        list, and sets the game to 'not ended'.
        """

        self.ended = False
        self.mistakes_count = 0
        self.mistakes_coord = list()

        self.grid = self.generate_level(self.grid_x, self.grid_y)


def pi_main() -> None:
    """
    This function is meant to be called once and just once, with the execution
    of the program.
    """

    game = PicrossGame(WIDTH, HEIGHT)

    gamelib.title(f"Simple Picross by NLGS")
    gamelib.resize(game.width, game.height)
    gamelib.icon(PICROSS_ICON)

    while gamelib.loop(fps=60):

        if game.ended:

            break

        gamelib.draw_begin()
        pigraphics.draw_screen(game)
        gamelib.draw_end()

        event = gamelib.wait(gamelib.EventType.ButtonPress)

        game.process_click(event.x, event.y, event.mouse_button)

        game.advance_game()

if __name__ == "__main__":

    gamelib.init(pi_main)


"""
Generic module that contains some useful objects:

* Timer

* Button

* Menu
"""

class Timer:
    """
    Simple Timer with countdown properties.
    """

    def __init__(self, init_time: int, message: str='') -> None:
        """
        Initializes an instance of type 'Timer'.
        """
        self.initial_time = init_time
        self.current_time = init_time
        self.msg = message

    def __str__(self) -> str:
        """
        Returns a string with class information so it can be printed later.
        """
        return f"Initital Time: {self.initial_time} - Current Time: {self.current_time}{f'Message: {self.msg}' if self.msg != '' else ''}"

    def deduct(self, how_much: int) -> None:
        """
        Descends the countdown subtracting 'how_much' time from 'self.current_time'.
        """
        self.current_time -= how_much

    def reset(self) -> None:
        """
        Resets the timer to its original value ('self.initial_value').
        """
        self.current_time = self.initial_time
        self.msg = ''

    def is_zero_or_less(self) -> bool:
        """
        Returns 'True' if the current time of the Timer reaches zero (0) or further, and 'False' otherwise.
        """
        return self.current_time <= 0

    def change_message(self, new_message: str) -> None:
        """
        Changes the current message to a given new one.
        """
        self.msg = new_message

class Button:
    """
    Simple class which defines a box and its corner coordinates, along with a message.
    """

    def __init__(self, x1: int, y1: int, x2: int, y2: int, message: str=''):
        """
        Initializes an instance of type 'Button'.

        It should always be true that 'x1 <= x2 && y1 <= y2'. As such, if it is not
        the case, those variables are inverted.
        """

        if x1 > x2:

            x1, x2 = x2, x1

        if y1 > y2:

            y1, y2 = y2, y1

        self.x1, self.y1, self.x2, self.y2 = x1, y1, x2, y2
        self.msg = message

    def __str__(self) -> str:
        """
        Returns a string with class information so it can be printed later.
        """
        return self.msg

    def all_coord(self) -> tuple[int, int, int, int]:
        """
        Returns a tuple with all the coordiantes of its hitbox.
        """

        return self.x1, self.y1, self.x2, self.y2

    def upper_left(self) -> tuple[int, int]:
        """
        Returns the UPPER LEFT coordinates of its hitbox.
        """
        return self.x1, self.y1

    def upper_right(self) -> tuple[int, int]:
        """
        Returns the UPPER RIGHT coordinates of its hitbox.
        """
        return self.x1, self.y1

    def bottom_left(self) -> tuple[int, int]:
        """
        Returns the BOTTOM LEFT coordinates of its hitbox.
        """
        return self.x1, self.y2

    def bottom_right(self) -> tuple[int, int]:
        """
        Returns the BOTTOM RIGHT coordinates of its hitbox.
        """
        return self.x2, self.y2

    def center(self) -> tuple[int, int]:
        """
        Return the CENTER coordinates of its hitbox.
        """
        return ((self.x2 + self.x1) // 2), ((self.y2 + self.y1) // 2)

    def is_inside(self, x: int, y: int) -> bool:
        """
        Returns 'True' if some given coordinates are inside the hitbox
        of this button.
        """

        return (self.x1 <= x <= self.x2) and (self.y1 <= y <= self.y2)

class Menu:
    """
    Class meant to hold many buttons and organize them in pages.
    """

    def __init__(self, button_titles: list,
                 area_corners: tuple[int, int, int, int],
                 window_size: int, max_rows: int=4, how_many_columns: int=1,
                 space_between: int=10, parent_menu=None, force_button_resize: bool=False) -> None:
        """
        Initializes an instance of type 'Menu'.

        'button_titles' must be a non-empty tuple.

        'max_rows' cannot be an integer lower than 1.

        'area_corners' must be a tuple of exactly 4 integers as its values.

        'window_size' is a tuple that contains the value of a given WIDTH and HEIGHT
        of the game window.

        'parent_menu' is another instance of type 'Menu' that has this instance
        as its child.

        'force_button_resize' means if the menu must use all the space in the area
        it is specified, which can resize the buttons.
        """

        if button_titles == ():

            raise Exception("'button_titles' cannot be empty. Must be an iteration with names (strings) and must have a length of at least 1.")

        if max_rows < 1:

            raise Exception("'max_rows' must be an integer of 1 or higher.")

        if not len(area_corners) == 4:

            raise Exception(f"area_corners has {len(area_corners)}. It must have exactly 4 integers as values.")

        button_titles = (button_titles.split("-=trash_value=-") if isinstance(button_titles, str) else list(button_titles))

        buttons_len = len(button_titles)

        how_many_rows = ((buttons_len // how_many_columns) if any((how_many_columns == 1, buttons_len % how_many_columns == 0)) else (buttons_len // how_many_columns) + 1)

        if force_button_resize and how_many_rows < max_rows:

            max_rows = how_many_rows

        # Measures
        self.area_x1, self.area_y1, self.area_x2, self.area_y2 = area_corners
        self.max_columns = how_many_columns
        self.max_rows = max_rows

        x_space = (self.area_x2 - self.area_x1) // self.max_columns
        y_space = (self.area_y2 - self.area_y1) // self.max_rows

        # Pages-related calculations
        self.max_pages = (((how_many_rows // self.max_rows) + 1) if all((not how_many_rows == self.max_rows, not how_many_rows % self.max_rows == 0)) else how_many_rows // self.max_rows)
        self.current_page = 1

        # Menu-related
        self.parent = parent_menu

        # Special Buttons
        width, height = window_size
        self.pgup_button = Button((self.area_x2 + space_between), self.area_y1, self.area_x2 + (y_space // 2), (self.area_y1 + (y_space // 2)), "/\\")
        self.pgdn_button = Button((self.area_x2 + space_between), (self.area_y2 - (y_space // 2)), self.area_x2 + (y_space // 2), self.area_y2, "\/")
        self.return_button = Button(self.area_x1, self.area_y1 - (height // 20), self.area_x1 + (width // 20), self.area_y1 - space_between, '<')

        # Button Lists
        self.buttons = self.generate_buttons(button_titles, x_space, y_space, space_between)
        self.buttons_on_screen = self.update_buttons()

        # Timers
        self.press_cooldown = Timer(20)

    @classmethod
    def sub_menu(cls, button_titles: list, corners: tuple[int, int, int, int], how_many_cols: int=1, space: int=10):
        """
        ______________________________________________________________________

        button_titles: <list> --> [<str>, <str>, ... , <str>]

        area_corners: <tuple> --> (<int>, <int>, <int>, <int>)

        how_many_columns: <int>


        ---> <Menu>
        ______________________________________________________________________

        It creates an instance of type 'Menu', but with the symbols for some buttons
        changed.
        """
        sub = cls(button_titles, corners, how_many_columns=how_many_cols, space_between=space)

        sub.pgup_button.msg = '^'
        sub.pgdn_button.msg = 'v'

        return sub

    def generate_buttons(self, titles_list: list, x_space: int, y_space: int, space_between: int=0) -> list:
        """
        ______________________________________________________________________

        titles_list: <list> -->  [<str>, <str>, ... , <str>]

        x_space, y_space, space_between: <int>


        ---> <list> --> [<Button>, <Button>, ... , <Button>]
        ______________________________________________________________________

        Generate buttons based on the effective area of the menu and the 'self.button_titles' list.
        'space_between' determines how much dead space there is between each button in said area.
        """
        buttons_list = list()
        cols_counter = 0
        rows_counter = 0

        for title in titles_list:

            cols_counter %= self.max_columns
            rows_counter %= self.max_rows

            x1 = (cols_counter * x_space) + self.area_x1 + (0 if cols_counter == 0 else space_between // 2)
            x2 = ((cols_counter + 1) * x_space) + self.area_x1 - (0 if cols_counter == (self.max_columns - 1) else space_between // 2)
            y1 = (rows_counter * y_space) + self.area_y1 + (0 if rows_counter == 0 else space_between // 2)
            y2 = ((rows_counter + 1) * y_space) + self.area_y1 - (0 if rows_counter == (self.max_rows - 1) else space_between // 2)

            buttons_list.append(Button(x1, y1, x2, y2, title))

            cols_counter += 1

            if cols_counter % self.max_columns == 0: # Go to next row only if the currnet column is filled first

                rows_counter += 1

        return buttons_list

    def update_buttons(self, page: int=1) -> list:
        """
        ______________________________________________________________________

        page: <int>


        ---> <list> --> [<Button>, <Button>, ... , <Button>]
        ______________________________________________________________________

        Updates the buttons list if the menu changes pages.

        The page number must be between 1 and the max values for the pages.
        """
        if 1 > page or self.max_pages < page:

            raise Exception(f"Page number is {page}. It must be between 1 and {self.max_pages} inclusive.") 

        buttons_list = list()

        for i in range((page - 1) * self.max_columns * self.max_rows, page * self.max_columns * self.max_rows):

            if i < len(self.buttons):

                buttons_list.append(self.buttons[i])

        if self.current_page < self.max_pages:

            buttons_list.append(self.pgdn_button)
        
        if self.current_page > 1:

            buttons_list.append(self.pgup_button)

        if self.parent: # add return button only if it is the main menu or a sub menu

            buttons_list.append(self.return_button)

        return buttons_list

    def change_page(self, to_next: bool=True, forced: bool=False) -> None:
        """
        Changes the current page to the previous or next one, depending of the parameter 'to_next'.
        If the new page is outside of the number of pages, does nothing if 'forced' is False, otherwise it rotates between the pages.
        """
        if forced:

            new_page = (self.max_pages % self.current_page) + 1

        else:

            new_page = (self.current_page + 1 if to_next else self.current_page - 1)
        
        if 1 <= new_page <= self.max_pages:

            self.current_page = new_page
            self.buttons_on_screen = self.update_buttons(new_page)